@@ -11,6 +11,12 @@ from __future__ import print_function, unicode_literals
 import collections
 import itertools
 
+try:
+    range = xrange
+except:
+    pass
+
+
 def num_char_count_bits(ver )  :
     return 16 if (ver + 7) // 17 else 8
 
@@ -583,3 +589,20 @@ def _get_bit(x , i )  :
 
 class DataTooLongError(ValueError):
     pass
+
+
+def qr2svg(qr , border )  :
+    parts  = []
+    for y in range(qr.size):
+        sy = border + y
+        for x in range(qr.size):
+            if qr.modules[y][x]:
+                parts.append("M%d,%dh1v1h-1z" % (border + x, sy))
+    t = """\
+<?xml version="1.0" encoding="UTF-8"?>
+<svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="0 0 {0} {0}" stroke="none">
+<rect width="100%" height="100%" fill="#F7F7F7"/>
+<path d="{1}" fill="#111111"/>
+</svg>
+"""
+    return t.format(qr.size + border * 2, " ".join(parts))
